import requests

BASE = "http://127.0.0.1:5000/api/v1/anomaly_detection/"

#contains features for sample data
test_data = {
    "data": [
        [0.2, 0.3, 0.4],
        [0.5, 0.6, 0.7],
        [0.8, 0.9, 1.0]
    ],
    "threshold": 0.5
}

response = requests.post(BASE + "train_cosmo", json=test_data)
if response.ok:
    print("Testing successful:", response.json())
else:
    print("Testing failed:", response.status_code, response.text)
test_data = {
    "data": [
        [0.2, 0.3, 0.4],
        [0.5, 0.6, 0.7],
        [0.8, 0.9, 1.0]
    ]
}

response = requests.post(BASE + "test_cosmo", json=test_data)
if response.ok:
    print("Testing successful:", response.json())
else:
    print("Testing failed:", response.status_code, response.text)

streaming_data = {
    "data": [
        [0.3, 0.4, 0.5]
    ]
}

response = requests.post(BASE + "test_streaming", json=streaming_data)
if response.ok:
    print("Streaming test successful:", response.json())
else:
    print("Streaming test failed:", response.status_code, response.text)

# Sample data for nonstationary anomaly detection
nonstationary_data = {
    "data": [
        [0.2, 0.3, 0.4],
        [0.5, 0.6, 0.7],
        [0.8, 0.9, 1.0]
    ],
    "n": 1
}

# Post request to nonstationary_AD_cosmo endpoint
response = requests.post(BASE + "nonstationary_AD_cosmo", json=nonstationary_data)
if response.ok:
    print("Nonstationary AD Cosmo test successful:", response.json())
else:
    print("Nonstationary AD Cosmo test failed:", response.status_code, response.text)

#
# Test data for deviation detection
# deviation_test_data = {
#     "data": [
#         [0.2, 0.3, 0.4],
#         [0.5, 0.6, 0.7],
#         [0.8, 0.9, 1.0]
#     ],
#     "mu": 0.5,  # Mean of the data
#     "sigma": 0.1,  # Standard deviation of the data
#     "l1": 4,
#     "l2": 8,
#     "l3": 12
# }
#
# # Send a POST request to the deviation detection endpoint
# response = requests.post(BASE + "deviation_detection", json=deviation_test_data)
# if response.ok:
#     print("Deviation Detection Test Successful:", response.json())
# else:
#     print("Deviation Detection Test Failed:", response.status_code, response.text)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pandas as pd
import os as os
import numpy as np
import sys
import statistics as st
import scipy as sp
import scipy.stats as sps
from scipy.stats import kurtosis
from scipy.stats import skew

#%% Time Domian Class
class TimeDomain:
    
    # def __init__(self, data):
    #     self.data = data

    @classmethod
    def extract_features_streaming(cls, data):
        chunkified_data = data.to_numpy()
        rms = cls.get_rms(chunkified_data)
        mean = cls.get_mean(chunkified_data)
        var =  cls.get_variance(chunkified_data)
        crest = cls.get_crestfactor(chunkified_data)
        kurt = cls.get_kurtosis(chunkified_data)
        skew = cls.get_skewness(chunkified_data)
        return rms, mean, var, crest, kurt, skew
    @classmethod
    def extract_features(cls, data, window_len=20480):
        data_np = data.to_numpy()
        chunkified_data = cls.get_chunks(data_np, frame_size=window_len, frame_shift=window_len)
        rms = cls.get_rms(chunkified_data)
        mean = cls.get_mean(chunkified_data)
        var =  cls.get_variance(chunkified_data)
        crest = cls.get_crestfactor(chunkified_data)
        kurt = cls.get_kurtosis(chunkified_data)
        skew = cls.get_skewness(chunkified_data)
        return rms, mean, var, crest, kurt, skew

    @classmethod
    def get_chunks(cls, data, frame_size=100, frame_shift=50):
        # initialize the number of chunks defined.
        chunks = []
        data = cls.check1(data)
        #Check if the provided size of the input data is greater than frame_size
        if(data.shape[0]<frame_size):
            sys.exit('Oops !!! The length of input vector is smaller than the analysis window length')
        for j in range(0,len(data)-frame_size,frame_shift):
            chunks.append(data[j:j+frame_size,:])
        frames = np.array(chunks)
        return frames
    @classmethod
    def check1(cls, data):
        # Check if the provided input data is atleast an 1D array
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        if(data.ndim < 1):
            sys.exit('Oops !!! Error with the input vector')
        elif(data.ndim == 1):
            data = np.reshape(data,(data.shape[0],1))
        return data
    @classmethod
    def get_rms(cls, data):
        """
        Parameters
        ----------
            data (pandas.DataFrame): input data in the dataframe format
        Returns
        -------
            rms (numpy.ndarray): root mean squared values
        """
        if data.ndim == 3:
            rms = np.sqrt(np.mean(data**2, axis = 1))
        else:
            rms = np.sqrt(np.mean(data**2, axis = 0))
        return rms
    @classmethod
    def get_mean(cls,data):
        """
        Parameters
        ----------
            data (pandas.DataFrame): input data in the dataframe format
        Returns
        -------
            mean (numpy.ndarray): mean values
        """
        if data.ndim == 3:
            mean = np.mean(data, axis = 1)
        else:
            mean = np.mean(data, axis = 0)
        return mean
    @classmethod
    def get_variance(cls,data):
        """
        Parameters
        ----------
            data (pandas.DataFrame): input data in the dataframe format
        Returns
        -------
            v (numpy.ndarray): variance values
        """
        if data.ndim == 3:
            v = np.var(data, axis =1)
        else:
            v = np.var(data, axis =0)
        return v
        
    @classmethod  
    def get_crestfactor(cls,data):
        """
        Parameters
        ----------
            data (pandas.DataFrame): input data in the dataframe format
        Returns
        -------
            c (numpy.ndarray): crestfactor values
        """
        if data.ndim == 3:
            peaks = np.max(data, axis = 1)
            rms = cls.get_rms(data)
            c = np.divide(peaks, rms)
        else:
            peaks = np.max(data, axis = 0)
            rms = cls.get_rms(data)
            c = np.divide(peaks, rms)
        return c
    @classmethod  
    def get_kurtosis(cls,data):
        if data.ndim == 3:
            kurt = kurtosis(data, axis=1)
        else:
            kurt = kurtosis(data, axis=0)
        return kurt
        
    @classmethod   
    def get_skewness(cls,data):
        if data.ndim == 3:
            sk = skew(data, axis = 1)
        else:
            sk = skew(data, axis = 0)
        return sk
        
        
#%% Frequency Domain (stationary)
        
class FrequencyDomain:
    
    def __init__(self,data):
        self.data = data
        
    def get_cepstrumcoeffs(self,data):
        spectrum = np.fft.fft(data,axis=1)
        ceps_coeffs = np.fft.ifft(np.log(np.abs(spectrum))).real
        return ceps_coeffs

        
   
    # def get_spectralanalysis(self,data):
        
        
#%% Time - Frequency (Non Stationary )
        

class Nonstationary:
    
    
    def __init__(self,data):
        self.data = data
      
    #Emphirical Mode Decomposition
    # def get_emd(self, data):
        
    #Wavelet Packet Decomposition
    # def get_wpd(self, data):
     
        
        
        
    


import requests

# Base URL of your Flask app running locally
base_url = "http://127.0.0.1:5000/api/v1/feature_engineering/"

# Sample data for testing time domain features
time_domain_data = {
    "data": [
        [0.1, 0.2, 0.3],
        [0.4, 0.5, 0.6],
        [0.7, 0.8, 0.9]
    ]
}

# Sample data for testing frequency domain features
frequency_domain_data = {
    "data": [
        [0.1, 0.2, 0.3],
        [0.4, 0.5, 0.6],
        [0.7, 0.8, 0.9]
    ]
}

# Send POST request for time domain features
response = requests.post(base_url + "time_domain_features", json=time_domain_data)
print("Time Domain Response:", response.json())

# Send POST request for frequency domain features
response = requests.post(base_url + "frequency_domain_features", json=frequency_domain_data)
print("Frequency Domain Response:", response.json())



# -*- coding: utf-8 -*-

#%%-----------------Section1: Importing fuction -------------------------------#
# from IPython import get_ipython
# get_ipython().magic('reset -sf')
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
import os as os
import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter
import warnings
from sklearn.preprocessing import StandardScaler
from sklearn import preprocessing
from sklearn import linear_model
from tensorflow.keras.models import Model, Sequential, load_model
from tensorflow.keras.layers import Dense, Input, LSTM, Dropout
from tensorflow.keras.layers import RepeatVector,TimeDistributed
from tensorflow.keras import optimizers
# from tcn import TCN
# import mplcursors
from datetime import datetime
import tkinter as tk
from tkinter import filedialog
import seaborn as sns
import tensorflow as tf
from minisom import MiniSom
import scipy.signal

sns.set(style="whitegrid")
warnings.filterwarnings("ignore")
plt.close('all')
np.random.seed(203)


#%%------------Section-2: Important Function for conversion and prediction----#
# creating 3D input matrix with lookback(no. of rows to look back) 

global anomaly_count,count

class AutoEncoder:
	def __init__(self):
		pass

	def train(self, x,L1=100,L2=100,e_dim=2,a_func='relu',b_size=30,epochs=100):
	    """
	    Autoencoder training with 2 encoder and 2 decoder dense layers. \n
	    Parameters
	    ----------------
	    x,L1,L2,e_dim,a_func,b_size,epochs:
	        x : Raw input data \n
	        L1 : Dense layer 1 with neurons (default=100) \n
	        L2 : Dense layer 2 with neurons (default=100) \n
	        e_dim : Latent representation or encoding dimension (default=2)
	        a_func : Activation function (default=relu) \n
	        b_size : Batch size (default=30)
	        epochs : Number of interations (default =100)
	    Returns
	    -------
	    AE,model,scaler:
	        AE : Autoencoder history from the trained model \n
	        model : Autoencoder model \n
	        scaler : Scaler used for scaling input data
	    """ 
	    x_train,x_test=train_test_split(x,test_size=0.2)
	    scaler=preprocessing.StandardScaler().fit(x_train)
	    x_train=scaler.transform(x_train)
	    x_test=scaler.transform(x_test)
	    ncol = x_train.shape[1]
	    input_dim = Input(shape = (ncol, ))

	    # DEFINE THE ENCODER LAYERS
	    encoded1 = Dense(L1, activation = 'linear')(input_dim)
	    encoded2 = Dense(L2, activation = a_func)(encoded1)
	    encoded3 = Dense(e_dim, activation = a_func)(encoded2)

	    # DEFINE THE DECODER LAYERS
	    decoded1 = Dense(L2, activation = a_func)(encoded3)
	    decoded2 = Dense(L1, activation = a_func)(decoded1) 
	    decoded3 = Dense(ncol, activation = 'linear')(decoded2)

	    model = Model(inputs = input_dim, outputs = decoded3)
	    model.compile(optimizer = 'SGD', loss = 'mse',metrics=['accuracy'])
	    AE = model.fit(x_train, x_train, epochs = epochs, batch_size = b_size,
	                         shuffle = True,validation_data = (x_test, x_test)) 
	    training_loss = AE.history['loss']
	    test_loss = AE.history['val_loss']
	    plt.figure()
	    plt.plot(training_loss, 'r--')
	    plt.plot(test_loss, 'b-')
	    plt.legend(['Training Loss', 'Test Loss'])
	    plt.xlabel('Epoch')
	    plt.ylabel('Loss')
	    plt.title('encoding_dim=' + str(e_dim))
	    return AE,model,scaler

	def predict(self, model,scaler,data):
	    """
	    Predicting target data from autoencoder model. \n
	    Parameters
	    ----------------
	    model,scaler,data:
	        model : Pre trainedAutoencoder model\n
	        scaler : Scaler used for scaling input data \n
	        data : Data from csv/xlsx file \n
	    Returns
	    -------
	    RE:
	        RE : Reconstruction error from the target data \n
	    """
	    x=scaler.transform(data)
	    pred=model.predict(x)
	    RE=np.abs(pred-x)
	    return RE  

	def plot(self, model,scaler,data,c_n,title):
	    """
	    Autoencoder model output plot . \n
	    Parameters
	    ----------------
	    model,c_n,data,title:
	        model : Autoencoder model \n
	        scaler : Scaler used for scaling input data \n
	        data: Input data should be of dataframe (used as index in plot) \n
	        c_n : Column number \n
	        title : Title of plot \n
	    Returns
	    -------
	    qe:
	        qe : Computed quantisation error \n
	    """
	    plt.figure()
	    x=scaler.transform(data)
	    pred=model.predict(x)
	    ax1=plt.subplot(2,1,1)
	    plt.plot(data.index,x[:,c_n])
	    plt.plot(data.index,pred[:,c_n])
	    plt.legend(['Data','Reconstructed'])
	    plt.title(title,fontsize=16,fontweight='bold')
	    plt.xticks(rotation=0, ha='right')
	    date_form = DateFormatter("%d/%b/%y\n%H:%M") # date and year
	    ax1.xaxis.set_major_formatter(date_form)  
	    ax2=plt.subplot(2,1,2,sharex=ax1)
	    RE=np.abs(pred[:,c_n]-x[:,c_n])
	    plt.plot(data.index,RE,'k')
	    plt.title('Reconstruction error',fontsize=16,fontweight='bold')
	    plt.xticks(rotation=0, ha='right')
	    ax2.xaxis.set_major_formatter(date_form)
	    # mplcursors.cursor()
	    plt.tight_layout()

class SOM:
	def __init__(self):
		pass

	# SOM Algorithm
	def train(self, x,w1=50,w2=50,sigma=0.1,lr=0.5,n_iter=500):
	    """
	    Training Self Organising map for input dataset\n
	    Parameters
	    ----------------
	    x,w1,w2,sigma,l_rate,n_int:
	        x : Input data \n
	        w1 & w2 : Dimension of output window \n
	        sigma : Radius
	        lr : Learning rate
	        n_iter : Number of interations
	    Returns
	    -------
	    som,q_error,scaler:
	        som: self organinsing map model \n
	        q_error: quantisation error of predicted data\n
	        scaler: scaling input data \n
	    """
	    scaler = preprocessing.MinMaxScaler()
	    x = scaler.fit_transform(x)
	    som = MiniSom(w1, w2, x.shape[1], sigma=sigma, learning_rate=lr)
	    som.random_weights_init(x)
	    som.train_random(x, n_iter) 
	 
	    return som, scaler

	def predict(self, som,data,scaler):
	    """
	    Prediction from SOM model \n
	    Parameters
	    ----------------
	    som,data,scaler:
	        som : SOM trained model \n
	        data : Input data from file \n
	        scaler : Scaler used for scaling data
	    Returns
	    -------
	    q_error:
	        q_error : Quantisation error from SOM \n
	    """
	    x=scaler.transform(data)
	    error=som.quantization(x)
	    q_error=self.quant_error(x,error)
	    return q_error
	    
	def quant_error(self, x,qant):
	    """
	    Quantisation error calculated based on input and predicted data. \n
	    Parameters
	    ----------------
	    x,qant:
	        x : Input data \n
	        qant : predicted data from som model
	    Returns
	    -------
	    qe:
	        qe : Computed quantisation error \n
	    """
	    qe = [] # quantization error
	    for i in range(len(x)):
	        qe.append(np.linalg.norm(qant[i] - x[i])) 
	    return np.array(qe)



#flask blueprint
# ref : https://flask.palletsprojects.com/en/2.3.x/blueprints/
from flask import Blueprint, request, jsonify
import pandas as pd
from IntelliMaint.anomaly_detection import AnomalyDetection

# Define the blueprint with a URL prefix
anomaly_blueprint = Blueprint('anomaly_detection', __name__)

# Initialize the AnomalyDetection model
ad_model = AnomalyDetection()


@anomaly_blueprint.route('/deviation_detection', methods=['POST'])
def deviation_detection_api():
    # Parse JSON data from the request
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "No data provided"}), 400

    mu = data.get('mu')
    sigma = data.get('sigma')
    l1 = data.get('l1', 4)
    l2 = data.get('l2', 8)
    l3 = data.get('l3', 12)

    if mu is None or sigma is None:
        return jsonify({"error": "Mean (mu) and standard deviation (sigma) are required"}), 400

    df = pd.DataFrame(data['data'])
    z_scores, sigma_used = ad_model.deviation_detection(df, mu, sigma, l1, l2, l3)
    return jsonify({
        "z_scores": z_scores.tolist(),
        "sigma": sigma_used
    }), 200

@anomaly_blueprint.route('/train_cosmo', methods=['POST'])
def train_anomaly_detection():
    data = request.get_json()  # Get JSON from request
    if not data or 'data' not in data:
        return jsonify({"error": "No data provided"}), 400

    # Extract data and optional threshold from the request
    df = pd.DataFrame(data['data'])
    threshold = data.get('threshold', 0.6)

    # Train the model
    ad_model.train_cosmo(df, threshold)
    return jsonify({"message": "Model trained successfully."}), 200

@anomaly_blueprint.route('/test_cosmo', methods=['POST'])
def test_anomaly_detection():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "No data provided"}), 400

    df = pd.DataFrame(data['data'])
    strangeness, p_values = ad_model.test_cosmo(df)
    return jsonify({
        "strangeness": strangeness.tolist(),
        "p_values": p_values.tolist()
    }), 200

@anomaly_blueprint.route('/test_streaming', methods=['POST'])
def test_anomaly_detection_streaming():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "No data provided"}), 400

    df = pd.DataFrame(data['data'])
    strangeness, p_values, deviation = ad_model.test_cosmo_streaming(df)
    return jsonify({
        "strangeness": strangeness.tolist(),
        "p_values": p_values.tolist(),
        "deviation": deviation.tolist()
    }), 200


@anomaly_blueprint.route('/nonstationary_AD_cosmo', methods=['POST'])
def nonstationary_AD_cosmo_api():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "No data provided"}), 400

    n = data.get('n', 1)
    df = pd.DataFrame(data['data'])
    strangeness, deviation, pvalue = ad_model.nonstationary_AD_cosmo(df, n)
    return jsonify({
        "strangeness": strangeness,
        "deviation": deviation,
        "pvalue": pvalue
    }), 200



from IntelliMaint.rul_models import GPRDegradationModel
import numpy as np
from flask import Blueprint, request, jsonify

rul_blueprint = Blueprint('rul_models', __name__)

# Create a dictionary to store models (simulate a database for demo purposes)
gpr_models = {}


@rul_blueprint.route('/create', methods=['POST'])
def create_model():
    data = request.get_json()
    if not data or 'HI' not in data or 'failure_threshold' not in data:
        return jsonify({"error": "Health Indicator and failure threshold are required"}), 400

    HI = np.array(data['HI']).reshape(-1, 1)
    failure_threshold = data['failure_threshold']
    order = data.get('order', 1)  # default order is 1

    # Create a model ID
    model_id = len(gpr_models) + 1
    gpr_models[model_id] = GPRDegradationModel(HI, failure_threshold, order)

    return jsonify({"message": "Model created successfully", "model_id": model_id}), 201

@rul_blueprint.route('/predict/<int:model_id>', methods=['POST'])
def predict(model_id):
    data = request.get_json()
    if not data or 'X' not in data:
        return jsonify({"error": "Time steps are required for prediction"}), 400

    if model_id not in gpr_models:
        return jsonify({"error": "Model not found"}), 404

    X = np.array(data['X']).reshape(-1, 1)
    model = gpr_models[model_id]
    Yp, Vp, rul = model.predict(X)

    # Ensure all numpy types are converted to Python native types for JSON serialization
    Yp = Yp.tolist()
    Vp = [float(v) for v in Vp]  # converting each element in variance to float
    rul = int(rul) if rul is not None else 'Not applicable'  # converting rul to int, or setting as a string if not applicable

    return jsonify({
        "predicted_HI": Yp,
        "variance": Vp,
        "estimated_RUL": rul
    }), 200

@rul_blueprint.route('/update/<int:model_id>', methods=['POST'])
def update(model_id):
    data = request.get_json()
    if not data or 'X' not in data or 'Y' not in data:
        return jsonify({"error": "Time steps and health indicator values are required"}), 400

    if model_id not in gpr_models:
        return jsonify({"error": "Model not found"}), 404

    X = np.array(data['X']).reshape(-1, 1)
    Y = np.array(data['Y']).reshape(-1, 1)
    model = gpr_models[model_id]
    model.update(X, Y)

    return jsonify({"message": "Model updated successfully"}), 200



"""Provides ways to transform the data of each system, so that the data from
various systems is more fairly comparable.
"""

__author__ = "Mohamed-Rafik Bouguelia"
__license__ = "MIT"
__email__ = "mohamed-rafik.bouguelia@hh.se"

import numpy as np
from scipy.stats import linregress
from .. import utils


class Transformer:
    def __init__(self, w=20, transformer="pvalue"):
        self.w = w
        self.transformer = transformer
        self.X = None
        self.P = None
        self.MIN_HISTORY_SIZE = 10

        # TODO: validate the parameters (accepted strings) for self.transformer

    def transform_all(self, X):
        Z = np.array([self.transform(x) for x in X])
        return Z

    def transform(self, x):
        if len(x) == 0 or self.transformer is None:
            return x

        if self.X is None and self.P is None:
            self.X = np.empty((0, len(x)))
            self.P = np.empty((0, len(x)))

        if self.transformer in ["mean_normalize", "std_normalize", "mean_std_normalize"]:
            with_mean = (self.transformer != "std_normalize")
            with_std = (self.transformer != "mean_normalize")
            return self.transform_normalize(x, with_mean, with_std)

        elif self.transformer in ["pvalue", "mean_pvalue"]:
            aggregate = (self.transformer == "mean_pvalue")
            return self.transform_pvalue(x, aggregate)

        elif self.transformer in ["slope"]:
            return self.transform_slope(x)

        else:
            raise utils.InputValidationError("transformer should be one of {}".format(["mean_normalize", "std_normalize", "mean_std_normalize", "slope", "pvalue", "mean_pvalue"]))


    def transform_slope(self, x):
        self.X = np.vstack([self.X, x])
        if len(self.X) < self.w: return np.zeros(x.shape)
        return np.array([ linregress(range(self.w), self.X[-self.w:, j])[0] for j in range(len(x)) ])


    def transform_normalize(self, x, with_mean=True, with_std=True):
        self.X = np.vstack([self.X, x])
        x_normalized = np.array(x)

        if with_mean:
            x_normalized -= np.mean(self.X, axis=0)

        if with_std:
            std = np.std(self.X, axis=0)
            x_normalized /= (std if std>0 else 1)

        return x_normalized

    def transform_pvalue(self, x, aggregate=True):
        if len(self.X) < self.MIN_HISTORY_SIZE:
            p_arr = [0.5 for _ in range(len(x))]
        else:
            pvalue = lambda val, values: len([1 for v in values if v > val]) / len(values)
            p_arr = [pvalue(x[i], self.X[:, i]) for i in range(len(x))]

        self.P = np.vstack([self.P, p_arr])
        self.X = np.vstack([self.X, x])

        z = []
        for i in range(len(x)):
            pvalues = self.P[-self.w:, i]
            if aggregate:
                z += [1. - np.mean(pvalues)] if len(pvalues) >= 1 else [0.5]
            else:
                z += [1. - pvalues[-1]] if len(pvalues) >= 1 else [0.5]

        return z



import numpy as np
import pandas as pd
from scipy.stats import kurtosis, skew
from scipy.fft import fft

def extract_features(signal, sampling_rate, bearing_info):
    # Time-domain features
    mean = np.mean(signal)
    std = np.std(signal)
    rms = np.sqrt(np.mean(signal**2))
    peak = np.max(np.abs(signal))
    skewness = skew(signal)
    kurtosis_value = kurtosis(signal)
    crest_factor = peak / rms
    shape_factor = rms / np.mean(np.abs(signal))
    impulse_factor = peak / np.mean(np.abs(signal))
    margin_factor = peak / (np.mean(np.sqrt(np.abs(signal))))**2

    # FFT features
    fft_signal = np.abs(fft(signal))
    fft_freq = np.fft.fftfreq(len(signal), 1/sampling_rate)
    fft_peaks = fft_signal[:len(signal)//2]
    fft_freqs = fft_freq[:len(signal)//2]

    # Bearing fault frequency features
    bpfo = bearing_info['bpfo']
    bpfi = bearing_info['bpfi']
    ftf = bearing_info['ftf']
    bsf = bearing_info['bsf']

    bpfo_amp = fft_signal[np.argmin(np.abs(fft_freq - bpfo))]
    bpfi_amp = fft_signal[np.argmin(np.abs(fft_freq - bpfi))]
    ftf_amp = fft_signal[np.argmin(np.abs(fft_freq - ftf))]
    bsf_amp = fft_signal[np.argmin(np.abs(fft_freq - bsf))]

    # Combine features into a dictionary
    features = {
        'mean': mean,
        'std': std,
        'rms': rms,
        'peak': peak,
        'skewness': skewness,
        'kurtosis': kurtosis_value,
        'crest_factor': crest_factor,
        'shape_factor': shape_factor,
        'impulse_factor': impulse_factor,
        'margin_factor': margin_factor,
        'fft_peaks': fft_peaks,
        'fft_freqs': fft_freqs,
        'bpfo_amp': bpfo_amp,
        'bpfi_amp': bpfi_amp,
        'ftf_amp': ftf_amp,
        'bsf_amp': bsf_amp
    }

    return features

# Example usage
bearing_info = {
    'bpfo': 3.5848,
    'bpfi': 5.4152,
    'ftf': 0.3983,
    'bsf': 4.7135
}

# Assuming you have a vibration signal and sampling rate
signal = ...  # Your vibration signal
sampling_rate = ...  # Sampling rate of the signal

features = extract_features(signal, sampling_rate, bearing_info)

# Convert features to a DataFrame
feature_df = pd.DataFrame([features])









import numpy as np
import os, glob
import re
import pandas as pd
import json

dirname = os.path.dirname(__file__)

class Utils:
	def __init__(self):
		self.turboengine_idx = 1
		self.bearing_idx = 4

	def get_features(self, component_type):

		def sorted_aphanumeric(data):
		    convert = lambda text: int(text) if text.isdigit() else text.lower()
		    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ] 
		    return sorted(data, key=alphanum_key)

		def state_of_health(data):
			initial_capacity = data[0]
			for i in range(len(data)):
				data[i] = (data[i]/initial_capacity)*100
			return data

		def get_battery_data():
			data_dir='examples/data/battery_data/B0006_discharge.json'
			data_dir = dirname + '/' + data_dir
			with open(data_dir) as f:    
			    discharge_data = json.load(f)
			data = []
			time_stamps = []
			for cycle in discharge_data.keys():
				data.append(discharge_data[cycle]["capacity"][0]) # remove the last dimension
				time_stamps.append(discharge_data[cycle]["date_time"])
			soh = np.array(state_of_health(data)).reshape(len(data), 1)
			times = np.array(time_stamps).reshape(len(time_stamps), 1)
			return soh

		if (component_type == 'turboengine'):
			data_dir='examples/data/phm08_data/csv'
			data_dir = dirname + '/' + data_dir
			dirlist = sorted_aphanumeric(os.listdir(data_dir))
			all_data = []
			for file in dirlist:
				if (not file.endswith('.py')) and (file != '__pycache__'):
					data = pd.read_csv(data_dir+"/"+file, engine='python')
					data = np.array(data)
					all_data.append(data[:, 2:])
			all_data_np = np.array(all_data)
			return all_data_np[self.turboengine_idx]
		elif (component_type == 'bearing'):
			data_dir='examples/data/bearing_data/'
			data_dir = dirname + '/' + data_dir
			rms = np.load(data_dir+'/3rd_test_bearing_'+str(self.bearing_idx)+'_rms.npy', allow_pickle=True)
			kurtosis = np.load(data_dir+'/3rd_test_bearing_'+str(self.bearing_idx)+'_kurtosis.npy', allow_pickle=True)
			crest = np.load(data_dir+'/3rd_test_bearing_'+str(self.bearing_idx)+'_crest.npy', allow_pickle=True)
			all_data_np = np.concatenate((rms, kurtosis, crest), axis=1)
			return all_data_np

		elif (component_type == 'battery'):
			return get_battery_data()
			
	def check1(self, data):
		assert isinstance(data,np.ndarray),'*** Please provide numpy array as input ***'
		assert len(data.shape)>0,'*** Ensure the array is not empty ***'
		assert np.all(np.isfinite(data)),'*** Remove NaN and Inf values from the numpy array'

	# Count the number of positive peaks
	def count_pos(self, data):
	    if(len(data.shape) == 1):
	        count = (data>0).sum()
	    else:
	        count = [(column>0).sum() for column in data.T]
	    return np.asarray(count)

	def count_neg(self, data):
	    if(len(data.shape) == 1):
	        count = (data<0).sum()
	    else:
	        count = [(column<0).sum() for column in data.T]
	    return np.asarray(count)


	def zscore(self, data, mu, sigma):
		z_s = (data - mu)/sigma
		return z_s    

	def zs_param(self, train, percentileScore=None):
		if(percentileScore == None):
			mu, sigma = np.mean(train), np.std(train)
		else:
			mu, sigma = np.percentile(train,percentileScore), np.std(train)
		return mu, sigma





from flask import Blueprint, request, jsonify
import pandas as pd
from IntelliMaint.feature_engineering import TimeDomain, FrequencyDomain

# Create a blueprint for feature engineering
feature_blueprint = Blueprint('feature_engineering', __name__)


@feature_blueprint.route('/time_domain_features', methods=['POST'])
def time_features_streaming():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "Data is required"}), 400

    try:
        df = pd.DataFrame(data['data'])

        # Add each feature calculation one at a time to isolate the problem
        rms = TimeDomain.get_rms(df.to_numpy())
        mean = TimeDomain.get_mean(df.to_numpy())
        var = TimeDomain.get_variance(df.to_numpy())
        crest = TimeDomain.get_crestfactor(df.to_numpy())
        kurt = TimeDomain.get_kurtosis(df.to_numpy())
        skew = TimeDomain.get_skewness(df.to_numpy())

        return jsonify({
            "rms": rms.tolist(),
            "mean": mean.tolist(),
            "variance": var.tolist(),
            "crest_factor": crest.tolist(),
            "kurtosis": kurt.tolist(),
            "skewness": skew.tolist()
        }), 200
    except Exception as e:
        return jsonify({"error": str(e)}), 500



@feature_blueprint.route('/frequency_domain_features', methods=['POST'])
def frequency_features():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "No data provided"}), 400

    try:
        # Convert data to DataFrame and then to numpy array
        df = pd.DataFrame(data['data'])
        array_data = df.to_numpy()

        # Create an instance of FrequencyDomain
        frequency_domain = FrequencyDomain(array_data)

        # Calculate cepstrum coefficients
        ceps_coeffs = frequency_domain.get_cepstrumcoeffs(array_data)

        return jsonify({
            "cepstrum_coefficients": ceps_coeffs.tolist()
        }), 200
    except Exception as e:
        return jsonify({"error": str(e)}), 500





# -*- coding: utf-8 -*-


import numpy as np
import matplotlib.pyplot as plt
# from .grand import IndividualAnomalyInductive, IndividualAnomalyTransductive, GroupAnomaly
from .grand.individual_anomaly.individual_anomaly_inductive import IndividualAnomalyInductive
from .grand.individual_anomaly.individual_anomaly_transductive import IndividualAnomalyTransductive
from .grand.group_anomaly.group_anomaly import GroupAnomaly
import pandas as pd
from scipy.stats import norm
from fpdf import FPDF

class AnomalyDetection:

    def __init__(self, w_martingale=15,k=50,non_conformity = "knn", ref_group=["hour-of-day"]):
        self.model = IndividualAnomalyTransductive(
                w_martingale = w_martingale,         # Window size for computing the deviation level
                ref_group = ref_group  # Criteria for reference group construction
                )

    def zscore(self, data, mu, sigma):
        return (data - mu) / sigma

    def deviation_detection(self, data, mu, sigma, l1 = 4, l2 = 8, l3 = 12):
        """
        Deviation detection using zscore \n
        Parameters
        ----------------
            data (pandas.DataFrame): input data in the form of dataframe
            mu (scaler): mean\n
            sigma (scaler): standard deviation\n
            l1: TO DO (default=4)
            l2: TO DO (default=8)
            l3: TO DO (default=12)
        Returns
        -------
            z_s (scaler): z-score
            sigma: standard deviation
        """
        z_s = self.zscore(data,mu, sigma)
        if(len(z_s.shape)>1):
            z_s = z_s[:,0]
        t = np.linspace(0,len(z_s)-1,len(z_s))
        thres1 = l1*sigma
        thres2 = l2*sigma
        thres3 = l3*sigma
        plt.scatter(t[np.where(z_s<=thres1)], z_s[np.where(z_s<=thres1)], color='y', label='Normal', alpha=0.3, edgecolors='none')
        plt.scatter(t[np.where((z_s>thres1) & (z_s<=thres2))], z_s[np.where((z_s>thres1) & (z_s<=thres2))], color='b', label='L1 Threshold', alpha=0.3, edgecolors='none')
        plt.scatter(t[np.where((z_s>thres2) & (z_s<=thres3))], z_s[np.where((z_s>thres2) & (z_s<=thres3))], color='g', label='L2 Threshold', alpha=0.3, edgecolors='none')
        plt.scatter(t[np.where(z_s>thres3)], z_s[np.where(z_s>thres3)], color='r', label='Anomalous points', alpha=0.3, edgecolors='none')
        plt.xlabel('Observation Signal (in samples)')
        plt.ylabel('Anomaly Score')
        plt.title('Anomaly Score Estimation')
        plt.legend()
        return z_s, sigma

    def train_cosmo(self,data, threshold=0.6, w_martingale = 15, non_conformity = "knn",k = 20):
        """
        Dummy text is this. Is it really though.\n
        Parameters
        ----------
            data (pandas.DataFrame): TO DO
            w_martingale (int): TO DO
            non_conformity (string): TO DO (default="median")
            k (int): TO DO (default=20)
        Returns
        -------
            model (object): trained model instance
        """
        df = data
        
        self.model = IndividualAnomalyInductive(
            w_martingale = w_martingale,
            non_conformity = non_conformity,
            k = k)

        # Fit the model to a fixed subset of the data
        X_fit = data.to_numpy()
        self.model.fit(X_fit)
        self.model.dev_threshold = threshold

    def test_cosmo(self, data):
        """
        TO DO some text for the time being.\n
        Parameters
        ----------
            data (pandas.DataFrame): TO DO
        Returns
        -------
            strangeness (pandas.DataFrame): TO DO
            P-values (pandas.DataFrame): TO DO
        """
        cols = ['Strangeness', 'P-Values', 'Deviation']
        lst_dict = []
        df = data
        for t, x in zip(df.index, df.values):
            info = self.model.predict(t, x)
            lst_dict.append({'Strangeness': info.strangeness,
                             'P-Values':info.pvalue,
                             'Deviation':info.deviation})
            
        # Plot strangeness and deviation level over time
        # gr = self.model.plot_deviations(figsize=(12, 8), plots=["strangeness", "deviation", "pvalue", "threshold"])
        
        # strangeness=0.0907913513823968, pvalue=0.2, deviation=0.26, is_deviating=False

        figsize=(12, 8)
        plots=["strangeness", "deviation", "pvalue", "threshold"]
        plots, nb_axs, i = list(set(plots)), 0, 0
        if "data" in plots:
            nb_axs += 1
        if "strangeness" in plots:
            nb_axs += 1
        if any(s in ["pvalue", "deviation", "threshold"] for s in plots):
            nb_axs += 1

        fig, axes = plt.subplots(nb_axs, sharex="row", figsize=figsize)
        if not isinstance(axes, (np.ndarray) ):
            axes = np.array([axes])

        if "data" in plots:
            axes[i].set_xlabel("Time")
            axes[i].set_ylabel("Feature 0")
            axes[i].plot(self.df.index, self.df.values[:, 0], label="Data")
            if debug:
                axes[i].plot(self.model.T, np.array(self.model.representatives)[:, 0], label="Reference")
            axes[i].legend()
            i += 1

        if "strangeness" in plots:
            axes[i].set_xlabel("Time")
            axes[i].set_ylabel("Strangeness")
            axes[i].plot(self.model.T, self.model.S, label="Strangeness")
            # if debug:
            #     axes[i].plot(self.model.T, np.array(self.model.diffs)[:, 0], label="Difference")
            axes[i].legend()
            i += 1

        if any(s in ["pvalue", "deviation", "threshold"] for s in plots):
            axes[i].set_xlabel("Time")
            axes[i].set_ylabel("Deviation")
            axes[i].set_ylim(0, 1)
            if "pvalue" in plots:
                axes[i].scatter(self.model.T, self.model.P, alpha=0.25, marker=".", color="green", label="p-value")
            if "deviation" in plots:
                axes[i].plot(self.model.T, self.model.M, label="Deviation")
            if "threshold" in plots:
                axes[i].axhline(y=self.model.dev_threshold, color='r', linestyle='--', label="Threshold")
            axes[i].legend()

        fig.autofmt_xdate()

        plt.savefig('cosmo.png')
        # pdf = FPDF()

        # Add a page
        # pdf.add_page()
        # pdf.image('cosmo.png')
        # pdf.output("test.pdf") 

        df1 = pd.DataFrame(lst_dict, columns=cols)
        
        return df1['Strangeness'].to_numpy(), df1['P-Values'].to_numpy()


    def test_cosmo_streaming(self, data):
        """
        TO DO some text for the time being.\n
        Parameters
        ----------
            data (pandas.DataFrame): TO DO
        Returns
        -------
            strangeness (pandas.DataFrame): TO DO
            P-values (pandas.DataFrame): TO DO
        """
        cols = ['Strangeness', 'P-Values', 'Deviation']
        lst_dict = []
        df = data
        for t, x in zip(df.index, df.values):
            info = self.model.predict(t, x)
            lst_dict.append({'Strangeness': info.strangeness,
                             'P-Values':info.pvalue,
                             'Deviation':info.deviation})
        df1 = pd.DataFrame(lst_dict, columns=cols)
        
        return df1['Strangeness'].to_numpy(), df1['P-Values'].to_numpy(), df1['Deviation'].to_numpy()
    
    def nonstationary_AD_cosmo(self,data,n=1):
        """
        TO DO Some text for the time being. \n
        Parameters
        ----------
        data (pandas.DataFrame): TO DO 
        n (int): TO DO
        w_martingale (int): TO DO 
        k (int): TO DO
        non_conformity (string): TO DO (default="median")
        ref_group (list(string)): TO DO (default=default=["hour-of-day"])

        Returns
        -------
        df1 (pandas.DataFrame): TO DO 
        gr (plot): TO DO
        """
        

        df = data
        cols = ['Strangeness', 'P-Values', 'Deviation']
        lst_dict = []
        
        info = self.model.predict(df.index[0], df.values[0])
    
        # lst_dict.append({'Strangeness': info.strangeness,
        #                      'P-Values':info.pvalue,
        #                      'Deviation':info.deviation})
        return info.strangeness, info.deviation, info.pvalue
        
        # result = []
        
        # for t, x in zip(df.index, df.values):
        #     info = self.model.predict(t, x)
    
        #     lst_dict.append({'Strangeness': info.strangeness,
        #                      'P-Values':info.pvalue,
        #                      'Deviation':info.deviation})
        #     result.append([info.strangeness, info.deviation, info.pvalue])
        # Plot strangeness and deviation level over time
        # gr = model.plot_deviations(figsize=(12, 8), plots=["strangeness", "deviation", "pvalue", "threshold"])
        
        # df1 = pd.DataFrame(lst_dict, columns=cols)
        
        # return df1, gr

        # return result


# Main Flask application
from flask import Flask

from flask_app.API_Blueprint_files.rul_api import rul_blueprint
from flask_app.API_Blueprint_files.anomaly_detection_api import anomaly_blueprint
from flask_app.API_Blueprint_files.feature_engineering_api import feature_blueprint
from flask_app.API_Blueprint_files.data_analysis_api import data_analysis_blueprint

def create_app():
    app = Flask(__name__)
    app.register_blueprint(anomaly_blueprint, url_prefix='/api/v1/anomaly_detection')
    app.register_blueprint(feature_blueprint, url_prefix='/api/v1/feature_engineering')
    app.register_blueprint(data_analysis_blueprint, url_prefix='/api/v1/data_analysis')
    app.register_blueprint(rul_blueprint, url_prefix='/api/v1/rul_models')
    return app

app = create_app()

@app.route('/')
def home():
    return "Welcome to the IntelliMaint API!"

if __name__ == "__main__":
    app.run(debug=True)



import pickle
from flask import Blueprint, request, jsonify
import pandas as pd
from tensorflow.keras.models import load_model

from IntelliMaint.data_analysis import AutoEncoder, SOM

data_analysis_blueprint = Blueprint('data_analysis', __name__)

@data_analysis_blueprint.route('/train_autoencoder', methods=['POST'])
def train_autoencoder():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "Data is required"}), 400

    try:
        x = pd.DataFrame(data['data'])
        L1 = data.get('L1', 100)
        L2 = data.get('L2', 100)
        e_dim = data.get('e_dim', 2)
        a_func = data.get('a_func', 'relu')
        b_size = data.get('b_size', 30)
        epochs = data.get('epochs', 100)

        ae = AutoEncoder()
        AE, model, scaler = ae.train(x, L1, L2, e_dim, a_func, b_size, epochs)

        # Save model and scaler for future use
        model.save('autoencoder_model.h5')
        pickle.dump(scaler, open('autoencoder_scaler.pkl', 'wb'))

        return jsonify({"message": "Autoencoder trained successfully"}), 200
    except Exception as e:
        return jsonify({"error": str(e)}), 500

@data_analysis_blueprint.route('/predict_autoencoder', methods=['POST'])
def predict_autoencoder():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "Data and model paths are required"}), 400

    try:
        df = pd.DataFrame(data['data'])
        model = load_model('autoencoder_model.h5')
        scaler = pickle.load(open('autoencoder_scaler.pkl', 'rb'))

        ae = AutoEncoder()
        RE = ae.predict(model, scaler, df)

        return jsonify({"reconstruction_error": RE.tolist()}), 200
    except Exception as e:
        return jsonify({"error": str(e)}), 500

@data_analysis_blueprint.route('/train_som', methods=['POST'])
def train_som():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "Data is required"}), 400

    try:
        df = pd.DataFrame(data['data'])
        w1 = data.get('w1', 50)
        w2 = data.get('w2', 50)
        sigma = data.get('sigma', 0.1)
        lr = data.get('lr', 0.5)
        n_iter = data.get('n_iter', 500)

        som = SOM()
        trained_som, scaler = som.train(df, w1, w2, sigma, lr, n_iter)

        # Save SOM and scaler for future use
        pickle.dump(trained_som, open('som_model.pkl', 'wb'))
        pickle.dump(scaler, open('som_scaler.pkl', 'wb'))

        return jsonify({"message": "SOM trained successfully"}), 200
    except Exception as e:
        return jsonify({"error": str(e)}), 500

@data_analysis_blueprint.route('/predict_som', methods=['POST'])
def predict_som():
    data = request.get_json()
    if not data or 'data' not in data:
        return jsonify({"error": "Data, SOM, and scaler paths are required"}), 400

    try:
        df = pd.DataFrame(data['data'])
        som = pickle.load(open('som_model.pkl', 'rb'))
        scaler = pickle.load(open('som_scaler.pkl', 'rb'))

        som_instance = SOM()
        q_error = som_instance.predict(som, df, scaler)

        return jsonify({"quantization_error": q_error.tolist()}), 200
    except Exception as e:
        return jsonify({"error": str(e)}), 500


# -*- coding: utf-8 -*-

from IntelliMaint.data_analysis import SOM
from IntelliMaint import Utils
import numpy as np
import pandas as pd

class HealthIndicator:
	"""

	This block provides the various metrics to identify the suitable Health indicators.
	Given a numpy matrix with each column representing some degrading features, this block provides the scores for 
	each features in relation to degradation

	score = computeHIScore(input, method_name = '')

	input should be a numpy array with finite numeric values.
	methods_name is a string with:
	mon - monotonicity
	pos_mon - positive monotonicity
	neg_mon - negative monotonocity

	This function returns the score, the higher the suitable that features are to represent the
	degradation characteristics. The score value will be between 0 and 1.

	"""
	def __init__(self, chunk_size=None, order=1):
		"""
		Parameters
		----------
		X -- Input features (prospective health indicators)
		algo --  Algorithms to check the preferred feature as health indicator
		chunk_size -- filtering window
		order -- order of the median filter

		Returns
		-------
			None
		"""
		self.chunk_size = chunk_size
		self.order = order

	def monotonicity(self, data):
		'''
		Monotonicity characterizes an increasing or decreasing trend. 
		It can be measured by the absolute difference of “positive” and “negative” derivatives for each feature.

		    Javed, Kamran, et al. "Enabling health monitoring approach based on vibration data for accurate prognostics." 
		    IEEE Transactions on Industrial Electronics 62.1 (2014): 647-656.

		Monotonicity will only the metrics with time \n
	
		Parameters
		----------
			data (pandas.DataFrame): input data in dataframe format
		Returns
		-------
			score (numpy.ndarray): measure of monotonicity
		'''
		utils = Utils()
		temp = np.diff(data, axis=0)
		score = np.abs(utils.count_pos(temp)-utils.count_neg(temp))/(data.shape[0]-1)
		return score

	def pos_monotonicity(self, data):
		'''
		Positive Prognosability: Characterizes an increasing trend. 
		Liao, Linxia, Wenjing Jin, and Radu Pavel. 
		"Enhanced restricted Boltzmann machine with prognosability regularization for prognostics and health assessment." 
		IEEE Transactions on Industrial Electronics 63.11 (2016): 7076-7083. \n

		Parameters
		----------
			data (pandas.DataFrame): input data in the dataframe format
		Returns
		-------
			score (numpy.ndarray): measure of the positive monotonicity
		'''
		utils = Utils()
		temp1 = np.diff(data, axis=0)
		temp2 = np.diff(data, 2, axis=0)
		score = (utils.count_pos(temp1)/(data.shape[0]-1) + utils.count_pos(temp2)/(data.shape[0]-2))/2
		return score

	    
	def neg_monotonicity(self, data):
		'''
		Negative Prognosability: Characterizes an decreasing trend. 
		Liao, Linxia, Wenjing Jin, and Radu Pavel. 
		"Enhanced restricted Boltzmann machine with prognosability regularization for prognostics and health assessment." 
		IEEE Transactions on Industrial Electronics 63.11 (2016): 7076-7083. \n

		Parameters
		----------
			data (pandas.DataFrame): input data in the dataframe format
		Returns
		-------
			score (numpy.ndarray): measure of the negative monotonicity
		'''
		utils = Utils()
		temp1 = np.diff(data, axis=0)
		temp2 = np.diff(data,2,axis=0)
		score = (utils.count_neg(temp1)/(data.shape[0]-1) + utils.count_neg(temp2)/(data.shape[0]-2))/2
		return score

	def linear_trendability(self, data):
		"""
		parameters
		----------
			data: input data (pandas.DataFrame)
		Returns
		-------
			scores: scores array with indices corresponding to each feature (np.numpy.ndarray)
		"""
		t = np.linspace(0,len(data)-1,num=len(data))
		if(len(data.shape) == 1):
			return  np.asarray(pearsonr(t,data)[0])
		t = np.linspace(0,len(data)-1,num=len(data))
		scores = np.asarray([pearsonr(t,column)[0] for column in data.T])
		return scores

	def nonlinear_trendability(self, data):
	    t = np.linspace(0,len(data)-1,num=len(data))
	    if(len(data.shape) == 1):
	        return np.asarray(spearmanr(t,data)[0])
	    return np.asarray([spearmanr(t,column)[0] for column in data.T])

	def computeScore(self, data, method='mon'):
	    # check1(data)
	    
	    if(method == 'mon'):
	        score = self.monotonicity(data)
	    elif(method == 'pos_mon'):
	        score = self.pos_monotonicity(data)
	    elif(method == 'neg_mon'):
	        score = self.neg_monotonicity(data)
	    elif(method == 'lin_trend'):
	        score = self.linear_trendability(data)
	    elif(method == 'nonlin_trend'):
	        score = self.nonlinear_trendability(data)
	    else:
	        print('Enter a valid option {\'mon\',\'pos_mon\',\'neg_mon\'}')
	    return score

	def computeHIScore(self, data, method='mon'):
	    if(self.chunk_size == None):
	        return self.computeScore(data, method)
	    if(self.order%2 == 0):
	        self.order = self.order + 1
	        data = signal.medfilt(data, self.order)
	    score = []
	    for i in range(0,len(data), self.chunk_size):
	        score.append(self.computeScore(input_data[i:i+self.chunk_size], method_name=method))
	    return np.asarray(score).reshape(1, len(score))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#%% Import Libraries
import os
import numpy as np 
import pandas as pd
import matplotlib.pyplot as plt
#plt.rcParams['figure.figsize'] = (10.0, 8.0)
import seaborn as sns
from scipy import stats
from scipy.stats import norm
from fpdf import FPDF
import IntelliMaint

package_dir_path = os.path.dirname(IntelliMaint.__file__)

#%% Exploratory Analysis Class
class ExploratoryAnalysis:
    
    @classmethod
    def perform_eda(cls, data):
        """
        Perform Exploratory Data Analysis.
        Parameters
        ----------
            data (pandas.DataFrame): input data in dataframe format
        Returns
        -------
            basic_stats_df (pandas.DataFrame): basic statistical info
            skewness_df (pandas.DataFrame): skewness values
            kurt_df (pandas.DataFrame): kurtosis values
        """
        pdf = FPDF()

        # Add a page
        pdf.add_page()
        pdf.set_font("Arial", size = 8)

        pdf.cell(200, 10, txt = "------------------------------------", 
        ln = 1, align = 'C')
        pdf.cell(200, 10, txt = "Performing Exploratory Data Analysis", 
        ln = 1, align = 'C')
        pdf.cell(200, 10, txt = "------------------------------------", 
        ln = 1, align = 'C')
        pdf.cell(200, 10, txt = "The train data has "+str(data.shape[0])+" rows and "+str(data.shape[1])+" columns", 
        ln = 1, align = 'C')
        
        print('------------------------------------')
        print("Performing Exploratory Data Analysis")
        print('------------------------------------')
        print("")
        #print("")
        print ('The train data has {0} rows and {1} columns'.format(data.shape[0], data.shape[1]))
        print("")

        
        def basic_stats():
            global basic_stats_df
            pdf.cell(200, 10, txt = "Descriptive Statistics", 
            ln = 1, align = 'C')
            pdf.cell(200, 10, txt = '----------------------', 
            ln = 1, align = 'C')
            print("Descriptive Statistics")
            print('----------------------')
            bs = data.describe()
            basic_stats_df = bs
            print(bs)
            pdf.cell(200, 10, txt = str(bs), 
            ln = 1, align = 'C')
            #print("")
            #return(bsf)
        
        def missing_value_plot():
            print("Missing Value Analysis")
            print('----------------------')
            pdf.cell(200, 10, txt = "Missing Value Analysis", 
            ln = 1, align = 'C')
            pdf.cell(200, 10, txt = '----------------------', 
            ln = 1, align = 'C')
            k = data.columns[data.isnull().any()]
            
            
            #visualising missing values
            if len(k) == 0:
                print("Hurray !!! No Missing Values")
                pdf.cell(200, 10, txt = "Hurray !!! No Missing Values", 
                ln = 1, align = 'C')
            else:
                print("Oops!! Some Values are missing, checkout the plot")
                pdf.cell(200, 10, txt = "Oops!! Some Values are missing, checkout the plot", 
                ln = 1, align = 'C')
                plt.figure()
                miss = data.isnull().sum()/len(data)
                miss = miss[miss > 0]
                miss.sort_values(inplace=True)
                miss = miss.to_frame()
                miss.columns = ['count']
                miss.index.names = ['Name']
                miss['Name'] = miss.index
    
                #plot the missing value count
                sns.set(style="whitegrid", color_codes=True)
                sns.barplot(x = 'Name', y = 'count', data=miss)
                plt.xticks(rotation = 90)
                #plt.show()
                
                
        def skewness_calc():
            global skewness_df
            print("")
            print('Skewness Analysis')
            print('-----------------')

            pdf.cell(200, 10, txt = 'Skewness Analysis', 
            ln = 1, align = 'C')
            pdf.cell(200, 10, txt = '----------------------', 
            ln = 1, align = 'C')

            sk = data.skew(axis = 0, skipna = True)
            sk = sk.to_frame()
            sk.columns = ["Skewness_Value"]
            skewness_df = sk
            print(sk)
            pdf.cell(200, 10, txt = str(sk), 
            ln = 1, align = 'C')
        def kurtosis_calc():
            global kurt_df
            print("")
            print('Kurtosis Analysis')
            print('-----------------')

            pdf.cell(200, 10, txt = 'Kurtosis Analysis', 
            ln = 1, align = 'C')
            pdf.cell(200, 10, txt = '----------------------', 
            ln = 1, align = 'C')

            kr = data.skew(axis = 0, skipna = True)
            kr = kr.to_frame()
            kr.columns = ["Kurtosis_Value"]
            kurt_df = kr
            print(kr)
            pdf.cell(200, 10, txt = str(kr), 
            ln = 1, align = 'C')

        def distribution_check():#assuming last column is the target variable
            #plt.figure()
            num = [f for f in data.columns if data.dtypes[f] != 'object']
            nd = pd.melt(data, value_vars = num)
            sns.set(font_scale=1)

            def dist_meanplot(x, **kwargs):
                sns.distplot(x, hist_kws=dict(alpha=0.3))
                plt.axvline(x.mean(),  color = 'red', label = str("mean: {:.2f}".format(x.mean())))
                plt.axvline(x.median(),linestyle ='--' , color = 'g',label = str("median: {:.2f}".format(x.median())))
                #plt.axvline(x.mode(),  color = 'o', label = str("mode: {:.2f}".format(x.mode())))

            g = sns.FacetGrid(nd, col="variable" ,col_wrap=2, sharex=False, sharey = False)
            g.map(dist_meanplot, "value").set_titles("{col_name}")
            for ax in g.axes.ravel():
                ax.legend(loc='best', fontsize = 'x-small')
            plt.tight_layout()
            plt.savefig('dist.png')
            # plt.show()    
            
        def correlation_check():
            plt.figure()
            corr = data.corr()
            mask = np.zeros_like(corr)
            mask[np.triu_indices_from(mask)] = True
            with sns.axes_style("white"):
                ax = sns.heatmap(corr,linewidths=.5, mask=mask, cmap="YlGnBu",annot=True)
                bottom, top = ax.get_ylim()
                ax.set_ylim(bottom + 0.5, top - 0.5)
                #ax.set_xlim(xmin=0.0, xmax=1000)

            plt.title('Correlation Graph')
            plt.tight_layout()
            # plt.show()
            plt.savefig('corr.png')
        

#        missing_values()
        #return(basic_stats()) 
        basic_stats()
        missing_value_plot()
        skewness_calc()
        kurtosis_calc()
        distribution_check()
        correlation_check()
        pdf.image('dist.png')
        pdf.image('corr.png')
        pdf.output("test.pdf") 
        return(basic_stats_df,skewness_df,kurt_df)


import numpy as np
from scipy.signal import hilbert, butter, lfilter
from scipy.stats import kurtosis
from scipy.fftpack import fft
import matplotlib.pyplot as plt

class SignalProcessor:
    def __init__(self, sampling_rate):
        self.sampling_rate = sampling_rate

    def ar_filter(self, signal, max_order):
        best_order = None
        max_kurtosis = -np.inf

        for order in range(1, max_order + 1):
            ar_coeffs = np.polyfit(signal[:-1], signal[1:], order)
            residual = signal[order:] - np.polyval(ar_coeffs, signal[:-order])
            k = kurtosis(residual)

            if k > max_kurtosis:
                max_kurtosis = k
                best_order = order

        ar_coeffs = np.polyfit(signal[:-best_order], signal[best_order:], best_order)
        residual = signal[best_order:] - np.polyval(ar_coeffs, signal[:-best_order])

        return residual
    
    def low_pass_filter(self, signal, cutoff_freq, order=5):
        nyquist_freq = 0.5 * self.sampling_rate
        normalized_cutoff_freq = cutoff_freq / nyquist_freq
        b, a = butter(order, normalized_cutoff_freq, btype='low', analog=False)
        filtered_signal = lfilter(b, a, signal)
        return filtered_signal

    def high_pass_filter(self, signal, cutoff_freq, order=5):
        nyquist_freq = 0.5 * self.sampling_rate
        normalized_cutoff_freq = cutoff_freq / nyquist_freq
        b, a = butter(order, normalized_cutoff_freq, btype='high', analog=False)
        filtered_signal = lfilter(b, a, signal)
        return filtered_signal

    def band_pass_filter(self, signal, low_cutoff_freq, high_cutoff_freq, order=5):
        nyquist_freq = 0.5 * self.sampling_rate
        normalized_low_cutoff_freq = low_cutoff_freq / nyquist_freq
        normalized_high_cutoff_freq = high_cutoff_freq / nyquist_freq
        b, a = butter(order, [normalized_low_cutoff_freq, normalized_high_cutoff_freq], btype='band', analog=False)
        filtered_signal = lfilter(b, a, signal)
        return filtered_signal

    def spectral_kurtosis(self, signal, window_lengths):
        num_windows = len(window_lengths)
        sk = []
        freqs = []

        for i, window_length in enumerate(window_lengths):
            window = np.hanning(window_length)
            overlap = window_length // 2

            num_segments = (len(signal) - window_length) // overlap + 1
            spec_kurtosis = np.zeros(window_length // 2 + 1)

            for j in range(num_segments):
                segment = signal[j * overlap : j * overlap + window_length]
                segment = segment * window
                spectrum = np.abs(np.fft.fft(segment))[:window_length // 2 + 1]
                spec_kurtosis += (spectrum**4) / (spectrum**2).mean()**2

            spec_kurtosis /= num_segments
            sk.append(spec_kurtosis)
            freqs.append(np.arange(window_length // 2 + 1) * self.sampling_rate / window_length)

        return sk, freqs

    def envelope_analysis(self, signal, bearing_frequencies):
        analytic_signal = hilbert(signal)
        envelope_signal = np.abs(analytic_signal)
        envelope_signal = envelope_signal - np.mean(envelope_signal)

        fft_envelope = np.abs(fft(envelope_signal)) / len(envelope_signal) * 2
        fft_envelope = fft_envelope[:len(fft_envelope)//2]

        freq_axis = np.arange(len(fft_envelope)) / len(envelope_signal) * self.sampling_rate

        plt.figure(figsize=(8, 4))
        plt.stem(freq_axis, fft_envelope, 'r', markerfmt=' ', basefmt='-b')
        plt.xlabel('Frequency (Hz)')
        plt.ylabel('Amplitude')
        plt.title('Envelope Spectrum')
        plt.xlim(0, 1000)

        for fault_freq in bearing_frequencies:
            plt.axvline(x=fault_freq, color='k', linestyle='--', linewidth=1)

        plt.show()




import requests

# Base URL of your Flask app running locally
base_url = "http://127.0.0.1:5000/api/v1/data_analysis/"

# Sample data for testing the AutoEncoder training
autoencoder_train_data = {
    "data": [
        [0.1, 0.2, 0.3, 0.4],
        [0.5, 0.6, 0.7, 0.8],
        [0.9, 1.0, 1.1, 1.2]
    ],
    "L1": 100,
    "L2": 100,
    "e_dim": 2,
    "a_func": "relu",
    "b_size": 10,
    "epochs": 10
}

# Send POST request to train the AutoEncoder
response = requests.post(base_url + "train_autoencoder", json=autoencoder_train_data,timeout=120)
print("Train AutoEncoder Response:", response.json())

# Prepare data for predicting using the trained AutoEncoder
autoencoder_predict_data = {
    "data": [
        [0.1, 0.2, 0.3, 0.4],
        [0.5, 0.6, 0.7, 0.8],
        [0.9, 1.0, 1.1, 1.2]
    ]
}

# Send POST request to predict using the AutoEncoder
response = requests.post(base_url + "predict_autoencoder", json=autoencoder_predict_data, timeout=120)
print("Predict AutoEncoder Response:", response.json())



import GPy
import numpy as np
# from skrvm import RVR
import matplotlib.pyplot as plt

class GPRDegradationModel:
	"""
	This block estimates remaining useful life of the components.
	"""
	def __init__(self, HI, failure_threshold, order=1):
		"""
		Parameters
		----------
			HI (numpy.ndarray): health indicator values array
			failure_threshold (scaler): the failure value of the health indicator
			order (int): order of the Gaussian Process kernel
		Returns
		-------
			None
		"""
		x = np.array([i for i in range(len(HI))]).reshape(-1, 1)
		y = HI
		print (x.shape, y.shape)
		self.kernel = GPy.kern.Poly(input_dim=HI.shape[1], variance=1., scale=1., bias=1., order=order)
		self.gpmodel = GPy.models.GPRegression(x, y, self.kernel)
		self.optimize()
		self.current_iteration = 0
		self.failure_threshold = failure_threshold

	def optimize(self):
		"""
		Parameters
		----------
			None
		Returns
		-------
			None
		"""
		self.gpmodel.optimize()

	def update(self, X, Y):
		"""
		Parameters
		----------
			X (np.array(int)): time steps array
			Y (np.array(float)): health indicator values array 

		Returns
		-------
			None
		"""
		self.gpmodel = GPy.models.GPRegression(X, Y, self.kernel)
		self.optimize()
		self.current_iteration += 1

	def predict(self, X):
		"""
		Parameters
		----------
			X (np.array(int)): ): time steps array

		Returns
		-------
			Yp (np.ndarray(float)): predicted health indicator values array
			Vp (np.ndarray(float)): variance in the prediction
			rul (float): remaining useful life
		"""
		Yp, Vp = self.gpmodel.predict(X)
		Yp, Vp = Yp.squeeze(), Vp.squeeze()
		failure_idx, = np.where(Yp >= self.failure_threshold)
		rul = None
		if (len(failure_idx) > 0):
			first_failure_idx = failure_idx[0]
			rul = first_failure_idx - self.current_iteration

		return Yp, Vp, rul

# class RVRDegradationModel:
# 	def __init__(self, HI):
# 		if (HI.shape[0] == 1):
# 			HI = HI.reshape(1, 1)
# 			timesteps = np.array([i for i in range(len(HI))]).reshape(len(HI), HI.shape[0])
# 		else:
# 			timesteps = np.array([i for i in range(len(HI))]).reshape(len(HI), HI.shape[1])
# 		self.rvrmodel = RVR(kernel='linear')
# 		self.optimize(timesteps, HI)

# 	def optimize(self, X, Y):
# 		self.rvrmodel.fit(X, Y)

# 	def update(self, X, Y):
# 		self.optimize(X, Y)

# 	def predict(self, X):
# 		# self.rvrmodel.fit(X, X)
# 		Yp = self.rvrmodel.predict(X)
# 		print (Yp)
# 		return Yp




from setuptools import setup, find_packages

# Read the contents of your README file
with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='IntelliMaint',
    version='0.12',
    author='IPTLP0032',
    author_email='iptgithub@intellipredikt.com',
    description='A prognostics package by IntelliPredikt Technologies',
    long_description=long_description, 
    long_description_content_type='text/markdown',  
    install_requires=[
        'scikit-learn', 'GPy', 'minisom', 'scipy', 'matplotlib', 'numpy>=1.16.1', 'mplcursors', 'fpdf2', 'tensorflow>=2.10,<2.12', 'keras>=2.10,<2.11', 'pandas>=1.0.0,<1.3.0', 'seaborn', 'imbalanced-learn',
    ],
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'IntelliMaint': [
            'grand/datasets/*',
            'grand/group_anomaly/*',
            'grand/individual_anomaly/*',
            'grand/*'
        ],
        'IntelliMaint.examples.data.battery_data': ['*'],
        'IntelliMaint.examples.data.bearing_data': ['*'],
        'IntelliMaint.examples.data.phm08_data.csv': ['*']
    },
    python_requires='>=3.5',   
)
